package com.jh.core.security;

import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.Locale;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.authentication.BadCredentialsException;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.AuthenticationException;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.StringUtils;

import com.jh.core.security.model.AppRole;
import com.jh.core.security.model.AppUser;
import com.jh.core.security.model.ResetPasswordToken;

@Service
public class AppUserAuthService implements AuthService {

	private static final Logger LOG = LoggerFactory.getLogger(AppUserAuthService.class);
	
	@Autowired
	private AppUserRepository userRepo;
	
	@Autowired
	private ResetPasswordRepository passwordRepo;
	
	@Autowired(required = false)
	private Collection<AuthServiceListener> authListeners;
	
	
	

	@Override
	@Transactional
	public Authentication authenticate(String login, String password) throws BadCredentialsException {
		AppUser u = userRepo.findByLogin(login);
		if (u == null) {
			throw new BadCredentialsException("AUTH.USER_NOT_FOUND");
		}
		if (!PasswordUtils.checkPassword(password, u.getPassword(), u.getPasswordSalt())) {
			throw new BadCredentialsException("AUTH.BAD_PASSWORD");
		}
		
		AppUser user = new AppUser();
		user.setAccountEnabled(u.isAccountEnabled());
		user.setAccountExpired(u.isAccountExpired());
		user.setAccountLocked(u.isAccountLocked());

		user.setDeleted(u.getDeleted());
		user.setEmail(u.getEmail());
		user.setFirstName(u.getFirstName());
		user.setId(u.getId());
		user.setLastName(u.getLastName());
		user.setLocale(u.getLocale());
		
		user.setLogin(u.getLogin());
		user.setRoles(new HashSet<>(u.getRoles()));
		
		//ADD implicit user
		user.getRoles().add(new AppRole("USER"));
		
		//initialize roles
		user.initRolesMap();
		
		Set<SimpleGrantedAuthority> authorities = u.getRoles().stream().map(r -> new SimpleGrantedAuthority("ROLE_" + r.getName())).collect(Collectors.toSet());
		AppUserAuthentication auth = new AppUserAuthentication(u, authorities);
		
		return auth;
	}

	@Override
	@Transactional
	public void register(String login, String email, String password, Locale locale) throws AuthenticationException {
		AppUser u = userRepo.findByLogin(login);
		if (u != null) {
			throw new BadCredentialsException("AUTH.USER_ALREADY_EXISTS");
		}
		if (!StringUtils.hasText(login) || !StringUtils.hasText(password)) {
			throw new BadCredentialsException("AUTH.BAD_USERNAME_OR_PASSWORD");
		}
		if (!PasswordUtils.validatePassword(password)) {
			throw new BadCredentialsException("AUTH.BAD_PASSWORD");
		}

		AppUser user = new AppUser();
		user.setLogin(login);
		user.setEmail(email);
		
		PasswordHash hash = PasswordUtils.hashPassword(password);
		user.setPassword(hash.getHash());
		user.setPasswordSalt(hash.getSalt());
		
		user.setAccountEnabled(true);
		user.setAccountExpired(false);
		user.setPasswordChanged(new Date());
		user.setPasswordExpired(false);

		userRepo.save(user);
		
		if (authListeners != null) {
			authListeners.forEach(al -> al.registerUser(user, password));
		}
	}

	@Override
	@Transactional
	public String generateResetToken(String login) {
		AppUser user = userRepo.findByLogin(login);
		if (user == null) {
			throw new BadCredentialsException("AUTH.USER_NOT_FOUND");
		}
		
		ResetPasswordToken token = new ResetPasswordToken(login, new Date(), UUID.randomUUID().toString());
		passwordRepo.save(token);
		
		if (authListeners != null) {
			authListeners.forEach(al -> al.generateResetToken(user, token));
		}
		
		return token.getToken();
	}

	
	@Override
	@Transactional
	public void resetPassword(String login, String token, String newPassword) throws AuthenticationException {
		ResetPasswordToken rpt = passwordRepo.findResetPasswordToken(login, token);
		if (rpt == null) {
			throw new BadCredentialsException("AUTH.INVALID_TOKEN");
		}
	
		AppUser user = userRepo.findByLogin(login);
		if (user == null) {
			throw new BadCredentialsException("AUTH.USER_NOT_FOUND");
		}
		
		PasswordHash hash = PasswordUtils.hashPassword(newPassword);
		user.setPassword(hash.getHash());
		user.setPasswordSalt(hash.getSalt());
		
		user.setPasswordChanged(new Date());
		user.setPasswordExpired(false);
		rpt.setUsed(true);
	}
	
	
	@Override
	@Transactional
	public void delete(AppUser user) throws AuthenticationException {
		if (user == null) {
			throw new BadCredentialsException("AUTH.USER_NOT_FOUND");
		}
		user = userRepo.findByLogin(user.getLogin());
		if (user == null) {
			throw new BadCredentialsException("AUTH.USER_NOT_FOUND");
		}
		
		user.setDeleted(true);
		user.setAccountEnabled(false);
	}
	
	@Override
	@Transactional
	public void changePassword(String login, String currentPassword, String newPassword) throws AuthenticationException {
		AppUser user = userRepo.findByLogin(login);
		if (user == null) {
			throw new BadCredentialsException("AUTH.USER_NOT_FOUND");
		}
		if (!PasswordUtils.checkPassword(currentPassword, user.getPassword(), user.getPasswordSalt())) {
			throw new BadCredentialsException("AUTH.BAD_PASSWORD");
		}
		
		PasswordHash hash = PasswordUtils.hashPassword(newPassword);
		user.setPassword(hash.getHash());
		user.setPasswordSalt(hash.getSalt());
	}
	
}
